import calendar
from datetime import timedelta

from enerdata.contracts.normalized_power import NormalizedPower
from enerdata.datetime.station import get_station
from enerdata.datetime.holidays import get_holidays
from enerdata.datetime.work_and_holidays import get_num_of_workdays_holidays


def check_range_hours(hours):
    before = (0, 0)
    for range_hours in sorted(hours):
        start, end = range_hours
        if start >= end:
            return False
        if start < 0 or start > 24:
            return False
        if end < 1 or end > 24:
            return False
        if start < before[1]:
            return False
        before = (start, end)
    return True


def are_powers_ascending(powers):
    power_ant = 0
    for power_new in powers:
        if power_new < power_ant:
            return False
        power_ant = power_new

    return True


class Tariff(object):
    """Energy DSO Tariff.
    """
    def __init__(self, code=None):
        self.code = code
        self._periods = tuple()
        self.cof = None
        self.require_powers_above_min_power = False

    @property
    def periods(self):
        return self._periods

    @periods.setter
    def periods(self, value):
        self._periods = value
        hours = {
            'holidays': [],
            'no_holidays': []
        }
        for period in self.energy_periods.values():
            if period.holiday:
                hours['holidays'].append(period)
            else:
                hours['no_holidays'].append(period)

        for station in ('summer', 'winter'):
            for holiday in (False, True):
                total_hours = 0
                range_hours = []
                for period in self.energy_periods.values():
                    if period.holiday == holiday:
                        total_hours += getattr(period, 'total_%s_hours' % station)
                        range_hours += getattr(period, '%s_hours' % station)
                range_hours = sorted(range_hours)
                if total_hours != 24:
                    if (holiday and total_hours) or not holiday:
                        raise ValueError(
                            'The sum of hours in %s%s must be 24h: %s'
                            % (station, holiday and ' (in holidays)' or '',
                            range_hours)
                        )
                if not check_range_hours(range_hours):
                    raise ValueError(
                        'Invalid range of hours in %s%s: %s'
                        % (station, holiday and ' (in holidays)' or '',
                           range_hours)
                    )

    @property
    def energy_periods(self):
        return dict((p.code, p) for p in self.periods if p.type == 'te')

    @property
    def power_periods(self):
        return dict((p.code, p) for p in self.periods if p.type == 'tp')

    def get_number_of_periods(self):
        return len([p for p in self.periods if p.type == 'te'])

    def get_min_power(self):
        return self.min_power

    def get_max_power(self):
        return self.max_power

    def get_period_by_date(self, date_time):
        datetime_previous_hour = date_time - timedelta(hours=1)
        station = get_station(datetime_previous_hour)
        date = datetime_previous_hour.date()
        holidays = get_holidays(date.year)
        if (calendar.weekday(date.year, date.month, date.day) in (5, 6) or
                date in holidays):
            holiday = True
        else:
            holiday = False

        # Map hour 0 to 24
        hour = date_time.hour or 24
        for period in self.periods:
            if period.holiday == holiday or not self.has_holidays_periods:
                for range_h in getattr(period, '%s_hours' % station):
                    if range_h[0] < hour <= range_h[1]:
                        return period
        return None


    @property
    def has_holidays_periods(self):
        return any(p.holiday for p in self.energy_periods.values())

    def is_maximum_power_correct(self, max_pow):
        return self.min_power < max_pow <= self.max_power

    def is_minimum_powers_correct(self, min_pow):
        if self.require_powers_above_min_power:
            return self.min_power < min_pow <= self.max_power
        else:
            return True

    @staticmethod
    def are_powers_normalized(powers):
        np = NormalizedPower()
        for power in powers:
            if not np.is_normalized(int(power * 1000)):
                return False

        return True

    def evaluate_powers(self, powers):
        if min(powers) <= 0:
            raise NotPositivePower()
        if not len(self.power_periods) == len(powers):
            raise IncorrectPowerNumber(len(powers), len(self.power_periods))
        if not self.is_maximum_power_correct(max(powers)):
            raise IncorrectMaxPower(max(powers), self.min_power, self.max_power)
        if not self.is_minimum_powers_correct(min(powers)):
            raise IncorrectMinPower(min(powers), self.min_power, self.max_power)
        if not self.are_powers_normalized(powers):
            raise NotNormalizedPower()

        return True

    def correct_powers(self, powers):
        raise NotImplementedError


class TariffPeriod(object):
    """Tariff period.
    """
    def __init__(self, code, ptype, **kwargs):
        assert ptype in ('te', 'tp')
        self.code = code
        self.type = ptype
        winter_hours = kwargs.pop('winter_hours', [(0, 24)])
        if not check_range_hours(winter_hours):
            raise ValueError('Invalid winter hours')
        self.winter_hours = winter_hours
        summer_hours = kwargs.pop('summer_hours', [(0, 24)])
        if not check_range_hours(summer_hours):
            raise ValueError('Invalid summer hours')
        self.summer_hours = summer_hours
        self.holiday = kwargs.pop('holiday', False)

    @property
    def total_summer_hours(self):
        hours = 0
        for range_h in self.summer_hours:
            hours += (range_h[1] - range_h[0])
        return hours

    @property
    def total_winter_hours(self):
        hours = 0
        for range_h in self.winter_hours:
            hours += (range_h[1] - range_h[0])
        return hours


class T20A(Tariff):
    def __init__(self):
        super(T20A, self).__init__()
        self.code = '2.0A'
        self.cof = 'A'
        self.periods = (
            TariffPeriod('P1', 'te'),
            TariffPeriod('P1', 'tp')
        )
        self.min_power = 0
        self.max_power = 10
        self.type = 'BT'

    def correct_powers(self, powers):
        try:
            if self.evaluate_powers(powers):
                return powers
        except:
            pass

        norm_power = NormalizedPower().get_norm_powers(
            int(self.min_power * 1000), int(self.max_power * 1000)
        ).next()

        return [norm_power / 1000.0] * len(self.power_periods)


class T20DHA(T20A):
    def __init__(self):
        super(T20DHA, self).__init__()
        self.code = '2.0DHA'
        self.cof = 'B'
        self.periods = (
            TariffPeriod(
                'P1', 'te',
                winter_hours=[(12, 22)],
                summer_hours=[(13, 23)]
            ),
            TariffPeriod(
                'P2', 'te',
                winter_hours=[(0, 12), (22, 24)],
                summer_hours=[(0, 13), (23, 24)]
            ),
            TariffPeriod(
                'P1', 'tp'
            )
        )


class T20DHS(T20DHA):
    def __init__(self):
        super(T20DHS, self).__init__()
        self.code = '2.0DHS'
        self.cof = 'D'
        self.periods = (
            TariffPeriod(
                'P1', 'te',
                winter_hours=[(13, 23)],
                summer_hours=[(13, 23)]
            ),
            TariffPeriod(
                'P2', 'te',
                winter_hours=[(0, 1), (7, 13), (23, 24)],
                summer_hours=[(0, 1), (7, 13), (23, 24)]
            ),
            TariffPeriod(
                'P3', 'te',
                winter_hours=[(1, 7)],
                summer_hours=[(1, 7)]
            ),
            TariffPeriod(
                'P1', 'tp'
            )
        )


class T21A(T20A):
    def __init__(self):
        super(T21A, self).__init__()
        self.code = '2.1A'
        self.cof = 'A'
        self.min_power = 10
        self.max_power = 15


class T21DHA(T20DHA):
    def __init__(self):
        super(T21DHA, self).__init__()
        self.code = '2.1DHA'
        self.cof = 'B'
        self.min_power = 10
        self.max_power = 15


class T21DHS(T20DHS):
    def __init__(self):
        super(T21DHS, self).__init__()
        self.code = '2.1DHS'
        self.cof = 'D'
        self.min_power = 10
        self.max_power = 15


class T30A(Tariff):
    def __init__(self):
        super(T30A, self).__init__()
        self.code = '3.0A'
        self.cof = 'C'
        self.min_power = 15
        self.max_power = 1000000
        self.require_powers_above_min_power = True
        self.type = 'BT'
        self.periods = (
            TariffPeriod(
                'P1', 'te',
                winter_hours=[(18, 22)],
                summer_hours=[(11, 15)]
            ),
            TariffPeriod(
                'P2', 'te',
                winter_hours=[(8, 18), (22, 24)],
                summer_hours=[(8, 11), (15, 24)]
            ),
            TariffPeriod(
                'P3', 'te',
                winter_hours=[(0, 8)],
                summer_hours=[(0, 8)]
            ),
            TariffPeriod(
                'P4', 'te',
                holiday=True,
                winter_hours=[(18, 22)],
                summer_hours=[(11, 15)]
            ),
            TariffPeriod(
                'P5', 'te',
                holiday=True,
                winter_hours=[(8, 18), (22, 24)],
                summer_hours=[(8, 11), (15, 24)]
            ),
            TariffPeriod(
                'P6', 'te',
                holiday=True,
                winter_hours=[(0, 8)],
                summer_hours=[(0, 8)]
            ),
            TariffPeriod(
                'P1', 'tp'
            ),
            TariffPeriod(
                'P2', 'tp'
            ),
            TariffPeriod(
                'P3', 'tp'
            )
        )


class T30ANoFestivos(T30A):
    def __init__(self):
        super(T30ANoFestivos, self).__init__()
        self.periods = (
            TariffPeriod(
                'P1', 'te',
                winter_hours=[(18, 22)],
                summer_hours=[(11, 15)]
            ),
            TariffPeriod(
                'P2', 'te',
                winter_hours=[(8, 18), (22, 24)],
                summer_hours=[(8, 11), (15, 24)]
            ),
            TariffPeriod(
                'P3', 'te',
                winter_hours=[(0, 8)],
                summer_hours=[(0, 8)]
            ),
            TariffPeriod(
                'P1', 'tp'
            ),
            TariffPeriod(
                'P2', 'tp'
            ),
            TariffPeriod(
                'P3', 'tp'
            )
        )

class T30A_one_period(T30A):
    """
    A 3.0A with one unique period
    """
    def __init__(self):
        super(T30A_one_period, self).__init__()
        self.periods = (
            TariffPeriod('P1', 'te'),
            TariffPeriod('P1', 'tp')
        )


class T31A(T30A):
    """
    3.1A Tariff patched with 3.1ALB
    """
    def __init__(self, kva=None):
        super(T31A, self).__init__()
        self.code = '3.1A'
        self.cof = 'C'
        self.min_power = 1
        self.max_power = 450
        self.losses = 0.04
        self.type = 'AT'
        self.require_powers_above_min_power = False
        if kva:
            if not isinstance(kva, (int, float)):
                raise ValueError('kva must be an enter value')
            self.low_voltage_measure = True
            self.kva = kva
        else:
            self.low_voltage_measure = False
        self.hours_by_period = {
            'P1': 6,
            'P2': 10,
            'P3': 8,
            'P4': 0,
            'P5': 6,
            'P6': 18
        }
        self.periods = (
            TariffPeriod(
                'P1', 'te',
                winter_hours=[(17, 23)],
                summer_hours=[(10, 16)]
            ),
            TariffPeriod(
                'P2', 'te',
                winter_hours=[(8, 17), (23, 24)],
                summer_hours=[(8, 10), (16, 24)]
            ),
            TariffPeriod(
                'P3', 'te',
                winter_hours=[(0, 8)],
                summer_hours=[(0, 8)]
            ),
            TariffPeriod(
                'P5', 'te',
                holiday=True,
                winter_hours=[(18, 24)],
                summer_hours=[(18, 24)]
            ),
            TariffPeriod(
                'P6', 'te',
                holiday=True,
                winter_hours=[(0, 18)],
                summer_hours=[(0, 18)]
            ),
            TariffPeriod(
                'P1', 'tp'
            ),
            TariffPeriod(
                'P2', 'tp'
            ),
            TariffPeriod(
                'P3', 'tp'
            )
        )

    @staticmethod
    def are_powers_normalized(powers):
        # 3.1A doesn't need to have normalized powers
        return True

    def apply_31A_LB_cof(self, balance, start_date, end_date):
        consumptions = balance.copy()
        period_hours = self.hours_by_period
        holidays_list = get_holidays(start_date.year)
        (workdays, holidays) = get_num_of_workdays_holidays(
            start_date, end_date, holidays_list
        )

        cofs = {}
        for period in period_hours:
            if period > 'P3':
                cofs[period] = period_hours.get(period, 0) * holidays
            else:
                cofs[period] = period_hours.get(period, 0) * workdays
        for period, consumption in balance.items():
            consumptions[period] = round(
                consumption * (1 + self.losses), 2
            ) + round(0.01 * cofs[period] * self.kva, 2)

        return consumptions

    def apply_curve_losses(self, measures):
        for idx, measure in enumerate(measures):
            values = measure._asdict()
            consumption = round(measure.measure * (1 + self.losses), 2) + round(0.01 * self.kva, 2)
            values['measure'] = consumption
            measures[idx] = measure._replace(**values)
        return measures

    def evaluate_powers(self, powers):
        super(T31A, self).evaluate_powers(powers)

        if not are_powers_ascending(powers):
            raise NotAscendingPowers()

        return True


class T31A_one_period(T31A):
    """
    A 3.1A with one unique period
    """
    def __init__(self):
        super(T31A_one_period, self).__init__()
        self.periods = (
            TariffPeriod('P1', 'te'),
            TariffPeriod('P1', 'tp')
        )


class T31ANoFestivos(T31A):
    def __init__(self, kva=None):
        super(T31ANoFestivos, self).__init__(kva=kva)
        self.hours_by_period = {
            'P1': 6,
            'P2': 10,
            'P3': 8,
        }
        self.periods = (
            TariffPeriod(
                'P1', 'te',
                winter_hours=[(17, 23)],
                summer_hours=[(10, 16)]
            ),
            TariffPeriod(
                'P2', 'te',
                winter_hours=[(8, 17), (23, 24)],
                summer_hours=[(8, 10), (16, 24)]
            ),
            TariffPeriod(
                'P3', 'te',
                winter_hours=[(0, 8)],
                summer_hours=[(0, 8)]
            ),
            TariffPeriod(
                'P1', 'tp'
            ),
            TariffPeriod(
                'P2', 'tp'
            ),
            TariffPeriod(
                'P3', 'tp'
            )
        )


class T61A(Tariff):
    """
    6.1A Tariff
    """
    def __init__(self):
        super(T61A, self).__init__()
        self.code = '6.1A'
        self.cof = 'C'
        self.min_power = 450
        self.max_power = 1000000
        self.type = 'AT'
        self.periods = (
            #
            # TODO: Implement correct energy periods
            #
            TariffPeriod('P1', 'te'),
            TariffPeriod(
                'P1', 'tp'
            ),
            TariffPeriod(
                'P2', 'tp'
            ),
            TariffPeriod(
                'P3', 'tp'
            ),
            TariffPeriod(
                'P4', 'tp'
            ),
            TariffPeriod(
                'P5', 'tp'
            ),
            TariffPeriod(
                'P6', 'tp'
            ),
        )

    @staticmethod
    def are_powers_normalized(powers):
        # 6.1A doesn't need to have normalized powers
        return True

    def evaluate_powers(self, powers):
        super(T61A, self).evaluate_powers(powers)

        if not are_powers_ascending(powers):
            raise NotAscendingPowers()

        return True


class T61B(T61A):
    """
    6.1B Tariff
    """
    def __init__(self):
        super(T61B, self).__init__()
        self.code = '6.1B'


class T62(T61A):
    """
    6.2 Tariff
    """
    def __init__(self):
        super(T62, self).__init__()
        self.code = '6.2'


class T63(T61A):
    """
    6.3 Tariff
    """
    def __init__(self):
        super(T63, self).__init__()
        self.code = '6.3'


class T64(T61A):
    """
    6.4 Tariff
    """
    def __init__(self):
        super(T64, self).__init__()
        self.code = '6.4'


class TRE(Tariff):
    def __init__(self):
        super(TRE, self).__init__()
        self.code = 'RE'
        self.cof = 'A'

        self.periods = (
            TariffPeriod('P0', 'te'),
            TariffPeriod('P0', 'tp')
        )


class NotPositivePower(Exception):
    def __init__(self):
        super(NotPositivePower, self).__init__(
            'Power should allways be higher than 0'
        )


class NotNormalizedPower(Exception):
    def __init__(self):
        super(NotNormalizedPower, self).__init__(
            'One or more of the powers doen\'t have a normalized value'
        )


class IncorrectPowerNumber(Exception):
    def __init__(self, power_number, expected_number):
        super(IncorrectPowerNumber, self).__init__(
            'Expected {0} power(s) and got {1}'.format(
                expected_number, power_number
            )
        )
        self.power_number = power_number
        self.expected_number = expected_number


class IncorrectMaxPower(Exception):
    def __init__(self, power, min_power, max_power):
        super(IncorrectMaxPower, self).__init__(
            'Power {0} is not between {1} and {2}'.format(
                power, min_power, max_power
            )
        )
        self.power = power
        self.min_power = min_power
        self.max_power = max_power


class IncorrectMinPower(Exception):
    def __init__(self, power, min_power, max_power):
        super(IncorrectMinPower, self).__init__(
            'Power {0} is not between {1} and {2}'.format(
                power, min_power, max_power
            )
        )
        self.power = power
        self.min_power = min_power
        self.max_power = max_power


class NotAscendingPowers(Exception):
    def __init__(self):
        super(NotAscendingPowers, self).__init__(
            'For this tariff powers should go in an ascending order (Pn+1>=Pn)'
        )


def get_tariff_by_code(code):
    """Get tariff class by code

    :param code: Tariff code
    :return: Tariff
    """
    available = {
        '2.0A': T20A,
        '2.0DHA': T20DHA,
        '2.0DHS': T20DHS,
        '2.1A': T21A,
        '2.1DHA': T21DHA,
        '2.1DHS': T21DHS,
        '3.0A': T30A,
        '3.0A C2': T30ANoFestivos,
        '3.1A': T31A,
        '3.1A LB': T31A,
        '3.1A C2': T31ANoFestivos,
        '6.1A': T61A,
        '6.1B': T61B,
        '6.2': T62,
        '6.3': T63,
        '6.4': T64,
        'RE': TRE,
    }
    return available.get(code, None)



from enerdata.contracts import TariffPeriod


class Measure(object):

    __slots__ = ('date', 'period', 'measure')

    def __init__(self, date, period, measure):
        self.date = date
        assert isinstance(period, TariffPeriod)
        self.period = period
        self.measure = measure

    def __lt__(self, other):
        return self.date < other.date

    def __le__(self, other):
        return self.date <= other.date

    def __gt__(self, other):
        return self.date > other.date

    def __ge__(self, other):
        return self.date >= other.date


class EnergyMeasure(Measure):

    __slots__ = ('type', 'consumption')

    def __init__(self, date, period, measure, mtype='A', consumption=0):
        assert period.type == 'te'
        super(EnergyMeasure, self).__init__(date, period, measure)
        self.type = mtype
        self.consumption = consumption

    def __repr__(self):
        return '<EnergyMeasure: [%s] %s - %s: %s>' % (
            self.date,
            self.period.code,
            self.period.type,
            self.measure
        )

    @classmethod
    def intervals(cls, measures):
        """Get the intervals of dates between measures.

        :param measures: A list of EnergyMeasures
        :return: A sorted list of `date` objects
        """

        dates = []
        for measure in sorted(measures):
            if measure.date not in dates:
                dates.append(measure.date)
        return dates


class PowerMeasure(Measure):
    def __init__(self, date, period, measure):
        assert period.type == 'tp'
        super(PowerMeasure, self).__init__(date, period, measure)


from datetime import date, timedelta


CONTRACT_STATES = [
    'draft',
    'validate',
    'pending',
    'active',
    'contract',
    'new_contract',
    'modcontract',
    'unpayed',
    'off',
    'down'
]


class Interval(object):
    def __init__(self, start, end, modification, changes):
        self.start = start
        self.end = end
        self.modification = modification
        self.changes = changes


class Contract(object):
    """Contract object.
    """
    def __init__(self):
        self.code = None
        self.state = 'draft'
        self.modifications = []
        self.contracted_power = 0

    def modify(self, modcontract, overwrite=False):
        assert isinstance(modcontract, Modification)
        if overwrite:
            self.modifications[-1] = modcontract
        else:
            if self.modifications:
                old_mod = self.modifications[-1]
                assert modcontract.start_date > old_mod.start_date
                old_mod.end_date = modcontract.start_date - timedelta(days=1)
                old_mod.next = modcontract
                modcontract.previous = old_mod
                old_mod.state = 'down'
                modcontract.version = old_mod.version + 1
            self.modifications.append(modcontract)
        self.apply_modification(modcontract)

    def get_modifications(self, start_date, end_date):
        mods = []
        # Get the modifications between dates
        for m in self.modifications:
            if m.start_date <= start_date and (
                    m.end_date is None or m.end_date >= end_date):
                mods.append(m)
            elif start_date <= m.start_date <= end_date:
                mods.append(m)
        return mods

    def get_intervals(self, start_date, end_date, changes=None):
        intervals = []
        previous_mod = None
        for m in self.get_modifications(start_date, end_date):
            diff = {}
            if previous_mod:
                diff = previous_mod.diff(m, changes)
            if diff or previous_mod is None:
                mod_end_date = m.end_date or end_date
                interval = Interval(
                    start=max(start_date, m.start_date),
                    end=min(mod_end_date, end_date),
                    changes=diff,
                    modification=m
                )
                intervals.append(interval)
            else:
                intervals[-1].end = min(m.end_date, end_date)
            previous_mod = m
        return intervals

    def get_changes(self, modification):
        changes = {}
        local_vals = vars(self)
        mod_vals = vars(modification)
        for attr, value in mod_vals.iteritems():
            if attr in local_vals:
                if value != local_vals[attr]:
                    changes[attr] = {'old': local_vals[attr], 'new': value}
        return changes

    def apply_changes(self, changes):
        for attr, value in changes.iteritems():
            setattr(self, attr, value['new'])

    def apply_modification(self, modification):
        changes = self.get_changes(modification)
        self.apply_changes(changes)


class Modification(object):
    def __init__(self, start_date=None):
        if start_date is None:
            start_date = date.today()
        self.start_date = start_date
        self.end_date = None
        self.previous = None
        self.next = None
        self.contracted_power = 0
        self.state = 'draft'
        self.version = 1

    def diff(self, modification, filter=None):
        changes = {}
        local_vals = vars(self)
        if modification is None:
            mod_vals = local_vals.copy()
        else:
            mod_vals = vars(modification)
        if filter is not None:
            mod_vals = dict((k, v) for k, v in mod_vals.items() if k in filter)
        for attr, value in mod_vals.iteritems():
            if attr in local_vals:
                if value != local_vals[attr]:
                    changes[attr] = {'old': local_vals[attr], 'new': value}
        return changes


# coding=utf-8
from workalendar.europe import Spain


class REECalendar(Spain):
    """Specific calendar for REE in Spain.

    The holidays are the same in Spain but only the ones which are not
    changeble and not fixed in a day.
    """
    include_epiphany = False
    include_good_friday = False





from enerdata.datetime import datetime
from enerdata.datetime.timezone import TIMEZONE
from pytz import utc


def convert_to_solar_hour(civil_hour):
    """
    Convert datetime to equivalent datetime solar hour
    :param civil_hour: datetimed
    :return: datetime equivalent solar hour
    """
    if not isinstance(civil_hour, datetime):
        raise ValueError('datetime was expected, found {}'.format(civil_hour))
    if not civil_hour.tzinfo:
        raise ValueError('datetime must be located: {}'.format(civil_hour))
    solar_hour = civil_hour.astimezone(utc)

    return solar_hour



from __future__ import division

import bisect
import logging
try:
    from collections import namedtuple, Counter
except ImportError:
    from backport_collections import namedtuple, Counter
from datetime import datetime, date, timedelta
from multiprocessing import Lock
from StringIO import StringIO
from dateutil.relativedelta import relativedelta
from decimal import Decimal

from enerdata.profiles import Dragger
from enerdata.contracts.tariff import (Tariff, T30A_one_period,
                                       T31A_one_period, T31A)
from enerdata.datetime.timezone import TIMEZONE
from enerdata.metering.measure import Measure, EnergyMeasure
from enerdata.datetime.holidays import get_holidays
from enerdata.datetime.work_and_holidays import get_num_of_workdays_holidays
from enerdata.datetime.solar_hour import convert_to_solar_hour

from os import path
import pandas as pd

logger = logging.getLogger(__name__)


class Coefficent(namedtuple('Coefficient', ['hour', 'cof'])):
    __slots__ = ()

    def __lt__(self, other):
        return self.hour < other.hour

    def __le__(self, other):
        return self.hour <= other.hour

    def __gt__(self, other):
        return self.hour > other.hour

    def __ge__(self, other):
        return self.hour >= other.hour


class Coefficients(object):
    def __init__(self, coefs=None):
        if coefs is None:
            coefs = []
        assert isinstance(coefs, list)
        self.coefs = list(coefs)

    def _check_pos(self, pos):
        if pos == len(self.coefs):
            raise ValueError('start date not found in coefficients')

    def insert_coefs(self, coefs):
        pos_0 = bisect.bisect_left(self.coefs, Coefficent(coefs[0][0], {}))
        pos_1 = bisect.bisect_right(self.coefs, Coefficent(coefs[-1][0], {}))
        logger.debug('Deleting from {start}({pos_0}) to {end}({pos_1})'.format(
            start=coefs[0][0], end=coefs[-1][0], pos_0=pos_0, pos_1=pos_1
        ))
        del self.coefs[pos_0:pos_1]
        for c in reversed(coefs):
            logger.debug('Inserting {c} into {pos_0}'.format(
                c=c, pos_0=pos_0
            ))
            self.coefs.insert(pos_0, c)

    def get(self, dt):
        assert isinstance(dt, datetime)
        if dt.dst() is None:
            dt = TIMEZONE.localize(dt)
        dt = TIMEZONE.normalize(dt)
        pos = bisect.bisect_left(self.coefs, Coefficent(dt, {}))
        self._check_pos(pos)
        return self.coefs[pos]

    def get_range(self, start, end):
        assert isinstance(start, date)
        assert isinstance(end, date)
        start = TIMEZONE.localize(datetime(
            start.year, start.month, start.day, 1), is_dst=True
        )
        # Sum one day to get the hour 00:00 of the next day
        end += timedelta(days=1)
        end = TIMEZONE.localize(datetime(
            end.year, end.month, end.day), is_dst=True
        ) + timedelta(seconds=1)
        pos = bisect.bisect_left(self.coefs, Coefficent(start, {}))
        self._check_pos(pos)
        end_pos = bisect.bisect_right(self.coefs, Coefficent(end, {}))
        return self.coefs[pos:end_pos]

    def get_coefs_by_tariff(self, tariff, start, end):
        assert hasattr(tariff, 'get_period_by_date')
        assert hasattr(tariff, 'energy_periods')
        assert isinstance(start, date)
        assert isinstance(end, date)
        sum_cofs = dict.fromkeys(tariff.energy_periods.keys(), 0)
        for hour, coef in self.get_range(start, end):
            if len(sum_cofs) > 1:
                period = tariff.get_period_by_date(hour)
                p_name = period.code
            else:
                p_name = sum_cofs.keys()[0]
            sum_cofs[p_name] += coef[tariff.cof]
        return sum_cofs


class Profiler(object):
    def __init__(self, coefficient):
        self.coefficient = coefficient

    def profile(self, tariff, measures, drag_method='hour'):
        """
        :param tariff:
        :param measures:
        :param drag_method: 'hour' means drag is passed to the next hour
                            'period' means drag is passed to the next hour for
                            the same period
        :return:
        """
        # {'PX': [(date(XXXX-XX-XX), 100), (date(XXXX-XX-XX), 110)]}
        _measures = list(measures)
        measures = {}
        for m in sorted(_measures):
            measures.setdefault(m.period.code, [])
            measures[m.period.code].append(m)
        measures_intervals = EnergyMeasure.intervals(_measures)
        logger.debug('Profiling {0} intervals'.format(len(measures_intervals)))
        for idx, measure_date in enumerate(measures_intervals):
            if idx + 1 == len(measures_intervals):
                break
            start = measure_date
            if idx > 0:
                start += timedelta(days=1)
            end = measures_intervals[idx + 1]
            logger.debug('Getting coeffs from {0} to {1}'.format(
                start, end
            ))
            sum_cofs = self.coefficient.get_coefs_by_tariff(tariff, start, end)
            dragger = Dragger()
            for hour, cof in self.coefficient.get_range(start, end):
                period = tariff.get_period_by_date(hour)
                if drag_method == 'hour':
                    dp = 'hour'
                else:
                    dp = period.code
                d = hour.date()
                if hour.hour == 0:
                    d -= timedelta(days=1)
                # To take the first measure
                if d == start:
                    d += timedelta(days=1)
                fake_m = Measure(d, period, 0)
                pos = bisect.bisect_left(measures.get(period.code, []), fake_m)
                pcode = period.code
                if pcode not in measures or pos >= len(measures[period.code]):
                    consumption = 0
                    consumption_date = None
                else:
                    consumption = measures[period.code][pos].consumption
                    consumption_date = measures[period.code][pos].date
                logger.debug('Hour: {0} Period: {1} Consumption: {2}'.format(
                    hour, period.code, consumption
                ))
                cof = cof[tariff.cof]
                hour_c = ((consumption * cof) / sum_cofs[period.code])
                aprox = dragger.drag(hour_c, key=dp)
                yield (
                    hour,
                    {
                        'aprox': aprox,
                        'drag': dragger[dp],
                        'consumption': consumption,
                        'consumption_date': consumption_date,
                        'sum_cofs': sum_cofs[period.code],
                        'cof': cof,
                        'period': period.code
                    }
                )


class REEProfile(object):
    HOST = 'www.ree.es'
    PATH = '/sites/default/files/simel/perff'
    down_lock = Lock()

    _CACHE = {}

    @classmethod
    def get_range(cls, start, end):
        cofs = []
        start = datetime(start.year, start.month, 1)
        end = datetime(end.year, end.month, 1)
        while start <= end:
            logger.debug('Downloading coefficients for {0}/{1}'.format(
                start.month, start.year
            ))
            cofs.extend(REEProfile.get(start.year, start.month))
            start += relativedelta(months=1)
        return cofs

    @classmethod
    def get(cls, year, month):
        try:
            import ssl
            try:
                _create_unverified_https_context = ssl._create_unverified_context
            except AttributeError:
                pass
            else:
                ssl._create_default_https_context = _create_unverified_https_context
        except ImportError:
            pass
        try:
            cls.down_lock.acquire()
            import csv
            import httplib
            key = '%(year)s%(month)02i' % locals()
            conn = None
            if key in cls._CACHE:
                logger.debug('Using CACHE for REEProfile {0}'.format(key))
                return cls._CACHE[key]
            perff_file = 'PERFF_%(key)s.gz' % locals()
            conn = httplib.HTTPSConnection(cls.HOST)
            conn.request('GET', '%s/%s' % (cls.PATH, perff_file))
            logger.debug('Downloading REEProfile from {0}/{1}'.format(
                cls.PATH, perff_file
            ))
            r = conn.getresponse()
            if r.msg.type == 'application/x-gzip':
                import gzip
                c = StringIO(r.read())
                m = StringIO(gzip.GzipFile(fileobj=c).read())
                c.close()
                reader = csv.reader(m, delimiter=';')
                header = True
                cofs = []
                for vals in reader:
                    if header:
                        header = False
                        continue
                    if int(vals[3]) == 1:
                        n_hour = 1
                    dt = datetime(
                        int(vals[0]), int(vals[1]), int(vals[2])
                    )
                    day = TIMEZONE.localize(dt, is_dst=bool(not int(vals[4])))
                    day += timedelta(hours=n_hour)
                    n_hour += 1
                    cofs.append(Coefficent(
                        TIMEZONE.normalize(day), dict(
                            (k, float(vals[i])) for i, k in enumerate('ABCD', 5)
                        ))
                    )
                cls._CACHE[key] = cofs
                return cofs
            else:
                raise Exception('Profiles from REE not found')
        finally:
            if conn is not None:
                conn.close()
            cls.down_lock.release()


class REProfile(object):
    translate_month = {
        1: 'Enero',
        2: 'Febrero',
        3: 'Marzo',
        4: 'Abril',
        5: 'Mayo',
        6: 'Junio',
        7: 'Julio',
        8: 'Agosto',
        9: 'Septiembre',
        10: 'Octubre',
        11: 'Noviembre',
        12: 'Diciembre'
    }

    @classmethod
    def get_range(cls, start, end):
        sheet_name = 'zona_{}'.format(cls.climatic_zone)
        filename = path.join(
            path.dirname(path.realpath(__file__)), 'data/coefficients_RE.xlsx'
        )
        df = pd.read_excel(filename, sheet_name=sheet_name)
        key = df.keys()[0]
        cofs = []
        while start <= end:
            month = cls.translate_month[start.month]
            solar_hour = convert_to_solar_hour(start)
            if solar_hour.hour != 0:
                hour = solar_hour.hour
            else:
                hour = 24
            coff_value = float(df[df[key] == month][hour])
            coff = Coefficent(start, {'A': coff_value})
            cofs.append(coff)
            start += relativedelta(hours=1)
        return cofs

    @classmethod
    def validate_exported_energy(cls, measures):
        """
        Check if there are ProfileHour with energy != 0 in periods with RE coefficient == 0
        :param measures: A list ProfileHour objects
        :return valid: A boolean that indicates if all the profiles in measures are valid
        :return invalid_profiles: A list containing the invalid profiles in measures
        """
        # Check if measures list is valid
        valid = True
        # Save invalid profiles
        invalid_profiles = []
        # get and iterate RE coefficients
        coefficients = cls.get_range(min(measures).date, max(measures).date)
        # iterate all RE measures
        for measure in measures:
            # if energy value is not zero
            if measure[1] != 0:
                # fetch coeff by profile date
                cof = False
                for c in coefficients:
                    if c.hour == measure[0]:
                        cof = c
                        break
                if cof:
                    # if coeff is zero
                    if cof.cof['A'] == 0:
                        # replace energy value by 'warning'
                        invalid_profiles.append(measure)
                        # mark measures as invalid
                        valid = False
        return valid, invalid_profiles


class REProfileZone1(REProfile):
    climatic_zone = 1


class REProfileZone2(REProfile):
    climatic_zone = 2


class REProfileZone3(REProfile):
    climatic_zone = 3


class REProfileZone4(REProfile):
    climatic_zone = 4


class REProfileZone5(REProfile):
    climatic_zone = 5


class REProfileHydraulic(REProfile):
    @classmethod
    def get_range(cls, start, end):
        filename = path.join(path.dirname(path.realpath(__file__)), 'data/coefficients_HIDRO_RE.csv')
        df = pd.read_csv(filename)
        cofs = []
        while start <= end:
            month = cls.translate_month[start.month]
            coff_value = float(df[df['MES'] == month]['Factor de funcionamiento'])
            coff = Coefficent(start, {'A': coff_value})
            cofs.append(coff)
            start += relativedelta(hours=1)
        return cofs


class REProfileFlat(REProfile):
    flat_cof = 0.85
    @classmethod
    def get_range(cls, start, end):
        cofs = []
        while start <= end:
            cofs.append(Coefficent(start, {'A': cls.flat_cof}))
            start += relativedelta(hours=1)
        return cofs


class ProfileHour(namedtuple('ProfileHour', ['date', 'measure', 'valid', 'accumulated'])):

    __slots__ = ()

    def __lt__(self, other):
        return self.date < other.date

    def __le__(self, other):
        return self.date <= other.date

    def __gt__(self, other):
        return self.date > other.date

    def __ge__(self, other):
        return self.date >= other.date


class Profile(object):
    """A Profile object representing hours and consumption.
    """

    def __init__(self, start, end, measures, accumulated=None, drag_by_periods=True):
        self.measures = measures[:]
        self.gaps = []  # Containing the gaps and invalid measures
        self.adjusted_periods = [] # If a period is adjusted
        self.start_date = start
        self.end_date = end
        self.profile_class = REEProfile

        assert type(drag_by_periods) == bool, "drag_by_periods must be a Boolean"
        self.drag_by_periods = drag_by_periods

        self.accumulated = Decimal(0)
        if accumulated:
            assert type(accumulated) == float or isinstance(accumulated, Decimal), "Provided accumulated must be a Decimal or a float"
            assert accumulated < 1 and accumulated > -1, "Provided accumulated '{}' must be -1 < accumulated < 1".format(accumulated)
            self.accumulated = accumulated

        measures_by_date = dict(
            [(m.date, m.measure) for m in measures if m.valid]
        )
        # End is included
        while start <= end:
            if measures_by_date.pop(TIMEZONE.normalize(start), None) is None:
                self.gaps.append(start)
            start += timedelta(hours=1)

    @property
    def n_hours(self):
        # End date is included, we have to sum one hour
        return int((self.end_date - self.start_date).total_seconds() / 3600) + 1

    @property
    def n_hours_measures(self):
        return len(self.measures)

    @property
    def total_consumption(self):
        return sum(x[1] for x in self.measures)

    @property
    def first_day_of_month(self):
        return self.end_date.day == 1 and self.end_date.hour > 0

    @staticmethod
    def simple_dragger(measures):
        dragger = Dragger()
        for idx, measure in enumerate(measures):
            values = measure._asdict()
            consumption = dragger.drag(measure.measure)
            values['measure'] = consumption
            measures[idx] = measure._replace(**values)
        return measures

    def get_hours_per_period(self, tariff, only_valid=False):
        assert isinstance(tariff, Tariff)
        hours_per_period = Counter()
        if only_valid:
            for m in self.measures:
                if m.valid:
                    period = tariff.get_period_by_date(m.date)
                    hours_per_period[period.code] += 1
        else:
            start_idx = self.start_date
            end = self.end_date
            while start_idx <= end:
                period = tariff.get_period_by_date(start_idx)
                hours_per_period[period.code] += 1
                start_idx += timedelta(hours=1)
        return hours_per_period

    def get_consumption_per_period(self, tariff):
        assert isinstance(tariff, Tariff)
        consumption_per_period = Counter()
        for period in tariff.energy_periods:
            consumption_per_period[period] = 0
        for m in self.measures:
            if m.valid:
                period = tariff.get_period_by_date(m.date)
                consumption_per_period[period.code] += m.measure
        return consumption_per_period

    def get_estimable_hours(self, tariff):
        assert isinstance(tariff, Tariff)
        total_hours = self.get_hours_per_period(tariff)
        valid_hours = self.get_hours_per_period(tariff, only_valid=True)
        estimable_hours = {}
        for period in total_hours.keys():
            estimable_hours[period] = total_hours[period] - valid_hours[period]
        return estimable_hours

    def get_estimable_consumption(self, tariff, balance):
        assert isinstance(tariff, Tariff)
        consumption_per_period = self.get_consumption_per_period(tariff)
        estimable = {}
        for period in consumption_per_period:
            estimable[period] = balance[period] - consumption_per_period[period]
        return estimable

    def estimate(self, tariff, balance):
        assert isinstance(tariff, Tariff)
        logger.debug('Estimating for tariff: {0}'.format(
            tariff.code
        ))

        # Adapt balance for simplified T30A with just one period
        if isinstance(tariff, T30A_one_period) or isinstance(tariff, T31A_one_period):
            balance = {
                "P1": sum([values for values in balance.values()])
            }
        # Adapt T31A6P adding P4 to P1
        if isinstance(tariff, T31A) and balance.get('P4', 0) > 0:
            balance['P1'] += balance['P4']
            balance['P4'] = 0

        measures = [x for x in self.measures if x.valid]
        start = self.start_date
        end = self.end_date
        # - REE cofs get from (year/month)
        # - Simel cofs get from (year/month/day hour) - can't substract one day
        if self.first_day_of_month or not issubclass(self.profile_class,
                                                     REEProfile):
            cofs = self.profile_class.get_range(start, end)
        else:
            cofs = self.profile_class.get_range(
                start, end - relativedelta(days=1)
            )
        cofs = Coefficients(cofs)
        cofs_per_period = Counter()

        for gap in self.gaps:
            period = tariff.get_period_by_date(gap)
            gap_cof = cofs.get(gap)
            cofs_per_period[period.code] += gap_cof.cof[tariff.cof]

        logger.debug('Coefficients per period calculated: {0}'.format(
            cofs_per_period
        ))

        energy_per_period = self.get_estimable_consumption(tariff, balance)
        energy_per_period_rem = energy_per_period.copy()

        dragger = Dragger()

        # Initialize the Dragger with passed accumulated value
        if len(self.gaps) > 0:
            # Drag by_hours
            if not self.drag_by_periods:
                init_drag_key = "default"
            else:
                init_drag_key = tariff.get_period_by_date(self.gaps[0]).code

            dragger.drag(self.accumulated, key=init_drag_key)

            for idx, gap in enumerate(self.gaps):
                logger.debug('Gap {0}/{1}'.format(
                    idx + 1, len(self.gaps)
                ))
                period = tariff.get_period_by_date(gap)

                drag_key = period.code if not self.drag_by_periods else "default"

                gap_cof = cofs.get(gap).cof[tariff.cof]
                energy = energy_per_period[period.code]
                # If the balance[period] < energy_profile[period] fill with 0
                if energy < 0:
                    energy = 0

                try:
                    gap_energy = (energy * gap_cof) / cofs_per_period[period.code]
                except ZeroDivisionError as error:
                    gap_energy = 0
                    logger.debug(error)

                aprox = dragger.drag(gap_energy, key=drag_key)
                energy_per_period_rem[period.code] -= gap_energy

                logger.debug(
                    'Energy for hour {0} is {1}. {2} Energy {3}/{4}'.format(
                        gap, aprox, period.code,
                        energy_per_period_rem[period.code], energy
                ))
                pos = bisect.bisect_left(measures, ProfileHour(gap, 0, True, 0.0))
                profile_hour = ProfileHour(TIMEZONE.normalize(gap), aprox, True, dragger[drag_key])

                measures.insert(pos, profile_hour)

        profile = Profile(self.start_date, self.end_date, measures)
        return profile

    def adjust(self, tariff, balance, diff=0):
        # Adjust values
        if self.gaps:
            raise Exception('Is not possible to adjust a profile with gaps')
        profile = Profile(self.start_date, self.end_date, self.measures)
        dragger = Dragger()
        energy_per_period = profile.get_consumption_per_period(tariff)
        for period_name, period_balance in balance.items():
            period_profile = energy_per_period[period_name]
            margin_bottom = period_balance - diff
            margin_top = period_balance + diff
            if not margin_bottom <= period_profile <= margin_top:
                profile.adjusted_periods.append(period_name)
                for idx, measure in enumerate(profile.measures):
                    period = tariff.get_period_by_date(measure.date).code
                    if period != period_name:
                        continue
                    values = measure._asdict()
                    values['valid'] = True
                    if not energy_per_period[period]:
                        values['measure'] = dragger.drag(measure.measure * 0)
                    else:
                        values['measure'] = dragger.drag(measure.measure * (
                            balance[period] / energy_per_period[period]
                        ))
                    profile.measures[idx] = measure._replace(**values)
        return profile

    def fixit(self, tariff, balance, diff=0):
        # Fill the gaps
        profile = self.estimate(tariff, balance)
        # Adjust to the balance
        profile = profile.adjust(tariff, balance, diff)

        if isinstance(tariff, T31A) and tariff.low_voltage_measure:
            # Apply losses
            profile.measures = tariff.apply_curve_losses(profile.measures)
            profile.measures = self.simple_dragger(profile.measures)
        return profile

    def __repr__(self):
        return '<Profile ({0} - {1}) {2}h {3}kWh>'.format(
            self.start_date, self.end_date, self.n_hours, self.total_consumption
        )



CHECKSUM_TABLE = 'TRWAGMYFPDXBNJZSQVHLCKE'


def checksum(number):
    try:
        rest0 = int(number) % 529
    except ValueError:
        raise
    cof = int(rest0 / 23)
    rest1 = rest0 % 23
    return CHECKSUM_TABLE[cof] + CHECKSUM_TABLE[rest1]


def check_cups_number(cups_number):
    tmp = cups_number[2:18]
    cs = checksum(tmp)
    check = 'ES%s%s' % (tmp, cs)
    if check != cups_number[:20]:
        return False
    return True


class CUPS(object):
    """CUPS object.
    """
    def __init__(self, number):
        if not check_cups_number(number):
            raise ValueError()
        self.number = number


import bisect
from datetime import date

from enerdata.metering.measure import *


class Meter(object):
    def __init__(self, serial, start_date=None):
        self.serial = serial
        self.energy_measures = []
        self.power_measures = []
        if not start_date:
            start_date = date.today()
        self.start_date = start_date
        self.end_date = None

    def add_measure(self, measure):
        assert isinstance(measure, (EnergyMeasure, PowerMeasure))
        assert measure.date >= self.start_date
        if self.end_date:
            assert measure.date <= self.end_date
        if isinstance(measure, EnergyMeasure):
            dst = self.energy_measures
        elif isinstance(measure, PowerMeasure):
            dst = self.power_measures
        pos = bisect.bisect_right(dst, measure)
        dst.insert(pos, measure)



import calendar
from datetime import timedelta


def get_num_of_workdays_holidays(init_date, end_date, holidays_list):
    workdays = 0
    holidays = 0

    _date = init_date
    while _date <= end_date:
        if (calendar.weekday(_date.year, _date.month, _date.day) in (5, 6)
        ) or (_date.date() in holidays_list):
            holidays += 1
        else:
            workdays += 1
        _date += timedelta(days=1)

    return workdays, holidays



import sys
from setuptools import setup, find_packages
from enerdata import __version__

PACKAGES_DATA = {'enerdata': ['profiles/data/*.xlsx', 'profiles/data/*.csv']}

INSTALL_REQUIRES = ['pytz', 'workalendar<8.0.0']

if sys.version_info < (2, 7):
    INSTALL_REQUIRES += ['backport_collections']

setup(
    name='enerdata',
    version=__version__,
    packages=find_packages(),
    url='http://code.gisce.net',
    license='MIT',
    author='GISCE-TI, S.L.',
    author_email='devel@gisce.net',
    install_requires=INSTALL_REQUIRES,
    package_data=PACKAGES_DATA,
    description='A set of energy models'
)



from enerdata.datetime import datetime
from enerdata.datetime.timezone import TIMEZONE


def get_station(dt):
    assert isinstance(dt, datetime)
    if not dt.tzinfo:
        dt = TIMEZONE.localize(dt)
    if TIMEZONE.normalize(dt).dst():
        return 'summer'
    else:
        return 'winter'
